@@ -1,5 +1,4 @@
 """Main module."""
-
 import argparse
 import io
 import os
@@ -12,6 +11,7 @@ import magic
 import stat
 import re
 import yaml
+import json
 import dataclasses as dc
 import dnf
 import datetime
@@ -19,6 +19,8 @@ import tarfile
 import hashlib 
 import time
 import glob
+import csv
+
 from wheel_filename import parse_wheel_filename
 
 from .utils import *
@@ -33,82 +35,26 @@ from pytictoc import TicToc
 t = TicToc()
 
 
-# import contextlib
-# import os
-# import distutils.dir_util
-
-# @contextlib.contextmanager
-# def monkeypatch(object, name, patch):
-#     value_orig = getattr(object, name)
-#     setattr(object, name, patch)
-#     yield object
-#     setattr(object, name, value_orig)
+def write_doc_table(filename, headers, rows):
+    with open(filename, 'w', encoding='utf-8') as lf:
+        lf.write(f"""
+<table class='wikitable' border=1>
+""")            
+        lf.write(f"""<tr>""")
+        for col_ in headers:
+            lf.write(f"""<th>{col_}</th>""")
+        lf.write(f"""</tr>\n""")
+        for row_ in rows:
+            lf.write(f"""<tr>""")
+            for col_ in row_:
+                lf.write(f"""<td>{col_}</td>""")
+            lf.write(f"""</tr>\n""")
+        lf.write(f"""
+</table>
+""")            
+    return
 
 
-# def copy_tree(src, dst, **kwargs):
-#     stdlib_symlink = os.symlink
-
-#     def _symlink(src, dst, **kwargs):
-#         try:
-#             stdlib_symlink(src, dst, **kwargs)
-#         except FileExistsError as err:
-#             pass
-
-#     with monkeypatch(distutils.dir_util.os, 'symlink', _symlink):
-#         distutils.dir_util.copy_tree(src, dst, **kwargs)
-
-
-# @dc.dataclass
-# class BinRegexps:
-#     '''
-#     Binary regexps. 
-#     '''
-#     need_patch: list #bins that need to be patched.   
-#     just_copy:  list #bins that just need to be copied.
-#     need_exclude:    dict #bins that just need to be copied.
-
-#     def __post_init__(self):
-#         self.just_copy_re = []
-#         if self.just_copy:
-#             for res_ in self.just_copy or []:
-#                 re_ = re.compile(res_ + '$')
-#                 self.just_copy_re.append(re_) 
-
-#         self.need_patch_re = []
-#         for res_ in self.need_patch or []:
-#             re_ = re.compile(res_ + '$')
-#             self.need_patch_re.append(re_) 
-
-#         self.need_exclude_re = {}
-#         for res_ in self.need_exclude.common or []:
-#             re_ = re.compile(res_ + '$')
-#             self.need_exclude_re[re_] = 0
-
-
-
-#     def is_just_copy(self, f):
-#         for re_ in self.just_copy_re:
-#             if re_.match(f):
-#                 return True
-#         return False    
-
-#     def is_need_patch(self, f):
-#         for re_ in self.need_patch_re:
-#             if re_.match(f):
-#                 return True
-#         return False    
-
-#     def is_need_exclude(self, f):
-#         for re_ in self.need_exclude_re:
-#             if re_.match(f):
-#                 self.need_exclude_re[re_] += 1
-#                 return True
-#         return False    
-
-#     def is_needed(self, f):
-#         return (self.is_just_copy(f) or self.is_need_patch(f)) and not self.is_need_exclude(f)
-    
-#     pass
 
 def fucking_magic(f):
     # m = magic.detect_from_filename(f)
@@ -246,6 +192,42 @@ class PythonPackages:
         return (self.build.projects or []) + (self.terra.projects or [])
 
 
+@dc.dataclass
+class GoPackagesSpec:
+    '''
+    Specification of set of go packages,
+    by goget modules and by git-nodes of some go projects
+    '''
+    # pip: list = None
+    projects: list = None
+
+
+@dc.dataclass
+class GoPackages:
+    '''
+        We separate go projects to two classes:
+        build: only needed for building on builder host
+        terra: needed to be install into terrarium
+    '''
+    build: GoPackagesSpec
+    terra: GoPackagesSpec
+
+    def __post_init__(self):    
+        '''
+        Recode from easydicts to objects
+        '''
+        self.build = GoPackagesSpec(**self.build)
+        self.terra = GoPackagesSpec(**self.terra)
+        pass
+
+    def projects(self):
+        '''
+        Get full list of go projects
+        ''' 
+        return (self.build.projects or []) + (self.terra.projects or [])
+
+
+
 @dc.dataclass
 class PackagesSpec:
     '''
@@ -324,6 +306,7 @@ class TerrariumAssembler:
             'build-wheels': 'compile wheels for our python sources',
             'install-wheels': 'Install our and external Python wheels',
             'build-nuitka': 'Compile Python packages to executable',
+            'build-go': 'Compile Go projects to executable',
             'make-isoexe': 'Also make self-executable install archive and ISO disk',
             'pack-me' :  'Pack current dir to time prefixed tar.bz2'
         }
@@ -355,6 +338,7 @@ class TerrariumAssembler:
             self.args.stage_build_wheels = True
             self.args.stage_install_wheels = True
             self.args.stage_build_nuitka = True
+            self.args.stage_build_go = True
             self.args.stage_pack = self.args.stage_build_and_pack
 
         if self.args.stage_my_source_changed:
@@ -364,6 +348,7 @@ class TerrariumAssembler:
             self.args.stage_build_wheels = True
             self.args.stage_install_wheels = True
             self.args.stage_build_nuitka = True
+            self.args.stage_build_go = True
             self.args.stage_pack = self.args.stage_my_source_changed
 
         if self.args.stage_download_all:
@@ -390,7 +375,9 @@ class TerrariumAssembler:
         except:
             pass # Building not on Fedora.        
 
-        self.spec = spec = yaml_load(specfile_, self.tvars)    
+        self.spec, vars_ = yaml_load(specfile_, self.tvars)    
+        self.tvars = edict(vars_)
+        spec = self.spec
 
         self.start_dir = os.getcwd()
          
@@ -430,6 +417,10 @@ class TerrariumAssembler:
 
         self.ps = PackagesSpec(**spec.packages)
         self.pp = PythonPackages(**spec.python_packages)
+        self.gp = None
+        if 'go_packages' in spec:
+            self.gp = GoPackages(**spec.go_packages)
+
         fs_ = []
         if 'folders' in spec:
             fs_ = spec.folders
@@ -525,7 +516,7 @@ set -x
             return
 
         tmpdir = os.path.join(self.curdir, "tmp/ta")
-        tmpdir_ = os.path.relpath(tmpdir)
+        tmpdir = os.path.relpath(tmpdir)
         bfiles = []
 
         #First pass
@@ -606,6 +597,46 @@ pipenv run python3 -m pip freeze > {target_dir_}/{build_name}-pip-freeze.txt
         pass
 
 
+    def build_go(self):
+        if not self.gp:
+            return
+
+        tmpdir = os.path.join(self.curdir, "tmp/ta")
+        bfiles = []
+
+        #First pass
+        module2build = {}
+        standalone2build = []
+        referenced_modules = set()
+
+        for td_ in self.gp.projects():
+            git_url, git_branch, path_to_dir_, _ = self.explode_pp_node(td_)
+            os.chdir(self.curdir)
+            if os.path.exists(path_to_dir_):
+                os.chdir(path_to_dir_)
+                path_to_dir__ = os.path.relpath(path_to_dir_, start=self.curdir)
+                outputname = os.path.split(path_to_dir_)[-1]
+                target_dir = os.path.join(tmpdir, outputname + '.build')
+                target_dir_ = os.path.relpath(target_dir, start=path_to_dir_)
+                lines = []
+                build_name = 'build_' + outputname
+                lines.append(fR"""
+pushd {path_to_dir_}       
+go mod download          
+CGO_ENABLED=0 go build -ldflags="-linkmode=internal -r" -o {target_dir_}/{outputname} 2>&1 >{outputname}.log
+popd
+    """ )
+                self.fs.folders.append(target_dir)
+                self.lines2sh(build_name, lines, None)
+                bfiles.append(build_name)
+
+        lines = []
+        for b_ in bfiles:
+            lines.append("./" + b_ + '.sh')
+        self.lines2sh("42-build-go", lines, "build-go")
+        pass
+
+
     def mycopy(self, src, dst):
         '''
         Адаптивная процедура копирования в подкаталоги окружения — симлинки релятивизируются 
@@ -760,37 +791,35 @@ pipenv run python3 -m pip freeze > {target_dir_}/{build_name}-pip-freeze.txt
                 lf.write('\n----------------\n')
                 lf.write(res)
     
+
         output  = subprocess.check_output(['repoquery'] + options_ + pl_,  universal_newlines=True).splitlines()
         output = [remove_epoch(x) for x in output if self.ps.is_package_needed(x)]
         packages_ = output + pl_
         with open(os.path.join(self.start_dir, 'selected-packages.txt'), 'w', encoding='utf-8') as lf:
             lf.write('\n- '.join(packages_))
 
-        with open(os.path.join(self.start_dir, 'selected-packages.rst'), 'w', encoding='utf-8') as lf:
-            lf.write("""
-+-------------------------------+-----------------+
-| Development tools             | Version         |
-+===============================+=================+
-| Visual Studio Code            |  1.47.0         |
-+-------------------------------+-----------------+""")            
-            packages_set_ = set()
-            for package_ in packages_:
-                purepackage = package_.split('.', 1)[0]
-                if len(purepackage) < len(package_):
-                    purepackage  = purepackage.rsplit('-', 1)[0]
-                packages_set_.add(purepackage)
 
-            for package_ in sorted(packages_set_):
-                res_ = list(self.installed_packages.filter(name=package_))
-                if len(res_)==0:
-                    continue
-                name_ = res_[0].name
-                version_ = res_[0].version
-                wtf = 1
-                lf.write("""
-|%(name_)-31s|%(version_)17s|
-+-------------------------------+-----------------+""" % vars())            
-                pass
+
+        packages_set_ = set()
+        for package_ in packages_:
+            purepackage = package_.split('.', 1)[0]
+            if len(purepackage) < len(package_):
+                purepackage  = purepackage.rsplit('-', 1)[0]
+            packages_set_.add(purepackage)
+
+        rows_ = []
+        for package_ in sorted(packages_set_):
+            res_ = list(self.installed_packages.filter(name=package_))
+            if len(res_)==0:
+                continue
+            name_ = res_[0].name
+            version_ = res_[0].version
+            wtf = 1
+
+            rows_.append([name_, version_])
+            pass
+
+        write_doc_table('doc-rpm-packages.htm', ['Packages', 'Version'], rows_)
 
         with open(cache_filename, 'w', encoding='utf-8') as lf:
             lf.write(','.join(packages_))
@@ -911,6 +940,18 @@ pipenv run python3 -m pip freeze > {target_dir_}/{build_name}-pip-freeze.txt
             #raise ex_
             pass
 
+    def projects(self):
+        """
+        return all projects list (Python/go/etc)
+        """
+        projects_ = []
+        if self.pp:
+            projects_ += self.pp.projects() 
+
+        if self.gp:
+            projects_ += self.gp.projects() 
+        return projects_
+
 
     def process_binary(self, binpath):
         '''
@@ -969,7 +1010,7 @@ pipenv run python3 -m pip freeze > {target_dir_}/{build_name}-pip-freeze.txt
         curdir = os.getcwd()
         args = self.args
         in_src = os.path.relpath(self.src_dir, start=self.curdir)
-        for td_ in self.pp.projects() + self.spec.templates_dirs:
+        for td_ in self.projects() + self.spec.templates_dirs:
             git_url, git_branch, path_to_dir_, _ = self.explode_pp_node(td_)
             os.chdir(curdir)
             if os.path.exists(path_to_dir_):
@@ -999,7 +1040,7 @@ mv in/src $snapshotdir
 mkdir -p {in_src}
 """)
         already_checkouted = set()
-        for td_ in self.pp.projects() + self.spec.templates_dirs:
+        for td_ in self.projects() + self.spec.templates_dirs:
             git_url, git_branch, path_to_dir_, _ = self.explode_pp_node(td_)
             if path_to_dir_ not in already_checkouted:
                 probably_package_name = os.path.split(path_to_dir_)[-1]
@@ -1205,10 +1246,12 @@ fi
 
     
         for rp_ in self.ps.repos or []:
-            if rp_.endswith('.rpm'):
+            if rp_.lower().endswith('.gpg'):
+                lines.append(f'sudo rpm --import {rp_} ')
+            elif rp_.endswith('.rpm'):
                 lines.append(f'sudo dnf install --nogpgcheck {rp_} -y ')
             else:
-                lines.append(f'sudo yum-config-manager --add-repo {rp_} -y ')
+                lines.append(f'sudo dnf config-manager --add-repo {rp_} -y ')
             pass
 
         self.lines2sh("00-install-repos", lines, "install-repos")    
@@ -1228,12 +1271,14 @@ fi
     
         lines = []
         lines_src = []
+        in_bin = os.path.relpath(self.in_bin, start=self.curdir)
 
+        scmd = f"rm -rf '{in_bin}/rpms'"
+        lines.append(scmd)
         scmd = "sudo yum-config-manager --enable remi"
         lines.append(scmd)
         pls_ = [p for p in self.need_packages + self.ps.build + self.ps.terra if isinstance(p, str)]
         purls_ = [p.url for p in self.need_packages + self.ps.build + self.ps.terra if not isinstance(p, str)]
-        in_bin = os.path.relpath(self.in_bin, start=self.curdir)
 
         packages = " ".join(self.dependencies(pls_, local=False) + purls_)    
         scmd = 'dnf download --skip-broken --downloaddir "%(in_bin)s/rpms" --arch=x86_64  --arch=x86_64 --arch=noarch  %(packages)s -y ' % vars()
@@ -1513,9 +1558,9 @@ rm -f *.tar.*
         parentdir, curname = os.path.split(self.curdir)
         disabled_suffix = curname + '.tar.bz2'
 
-        banned_ext = ['.old', '.iso', disabled_suffix]
+        banned_ext = ['.old', '.iso', '.lock', disabled_suffix, '.dblite', '.tmp', '.log']
         banned_start = ['tmp']
-        banned_mid = ['/out/', '/wtf/', '/.vagrant/', '/.git/']
+        banned_mid = ['/out', '/wtf', '/ourwheel/', '/.vagrant', '/.git', '/.vscode', '/key/', '/tmp/', '/src.', '/bin.',  '/cache_', 'cachefilelist_', '/.image', '/!']
 
         # there are regularly some files unaccessable for reading.
         self.cmd('sudo chmod a+r /usr/lib/cups -R')
@@ -1543,7 +1588,7 @@ rm -f *.tar.*
         tbzname = os.path.join(self.curdir, 
                 "%(time_prefix)s-%(curname)s.tar.bz2" % vars())
         tar = tarfile.open(tbzname, "w:bz2")
-        tar.add(self.curdir, recursive=True, filter=filter_)
+        tar.add(self.curdir, "./sources-for-audit", recursive=True, filter=filter_)
         tar.close()    
 
     def remove_exclusions(self):
@@ -1610,6 +1655,7 @@ rm -f *.tar.*
             
                 file_loader = FileSystemLoader(path_to_dir)
                 env = Environment(loader=file_loader)
+                env.filters["hash"] = j2_hash_filter    
                 env.trim_blocks = True
                 env.lstrip_blocks = True
                 env.rstrip_blocks = True            
@@ -1635,6 +1681,8 @@ rm -f *.tar.*
                         
                         plain = False
                         try:
+                            if 'users.xml' in fname_:
+                                dfdsfdsf=1
                             m = fucking_magic(fname_)
                             if 'text' in m:
                                 plain = True
@@ -1691,6 +1739,7 @@ rm -f *.tar.*
         self.build_wheels()
         self.install_wheels()
         self.build_nuitkas()
+        self.build_go()
         # self.install_packages()
 
         # if self.args.stage_build_nuitka:
@@ -1698,6 +1747,19 @@ rm -f *.tar.*
         # self.build_nuitkas()
             # return
 
+        try:
+            output_ = subprocess.check_output('pipenv run pip list --format json', shell=True)            
+            json_ = json.loads(output_)
+
+            rows_ = []
+            for r_ in json_:
+                rows_.append([r_['name'], r_['version']])
+
+            write_doc_table('doc-python-packages.htm', ['Package', 'Version'], sorted(rows_))
+        except Exception as ex_:
+            print(ex_)
+            pass    
+
         specfile_ = self.args.specfile
         self.lines2sh("50-pack", [
             '''
@@ -1730,6 +1792,22 @@ terrarium_assembler --debug --stage-pack=./out-debug "%(specfile_)s" --stage-mak
             root_dir = self.root_dir = expandpath(args.stage_pack)
             # self.remove_exclusions()
 
+            cloc_csv = 'tmp/cloc.csv'
+            if not os.path.exists(cloc_csv):
+                if shutil.which('cloc'):
+                    os.system('cloc ./in/src/ --csv  --report-file=tmp/cloc.csv --3')
+            if os.path.exists(cloc_csv):
+                table_csv = []
+                with open(cloc_csv, newline='') as csvfile:
+                    csv_r = csv.reader(csvfile, delimiter=',', quotechar='|')
+                    for row in list(csv_r)[1:]:
+                        row[-1] = int(float(row[-1]))
+                        table_csv.append(row)
+
+                table_csv[-1][-2], table_csv[-1][-1] = table_csv[-1][-1], table_csv[-1][-2]       
+                write_doc_table('doc-cloc.htm', ['Файлов', 'Язык', 'Пустых', 'Комментариев', 'Строчек кода', 'Мощность языка', 'COCOMO строк'], 
+                                table_csv)        
+
             # install_templates(root_dir, args)
 
             packages_to_deploy = []
@@ -1863,7 +1941,9 @@ terrarium_assembler --debug --stage-pack=./out-debug "%(specfile_)s" --stage-mak
                             self.add(f, libfile, recursive=False)
                     pass
 
-    
+
+
+
             install_templates(root_dir, args)
             self.install_terra_pythons()
             # install_templates(root_dir, args)
